'' Propeller Z-Machine 
'' This code is partially based on Zax 0.91 by Matthew Kimmel 
'' (c) 2010 Mikhail Gorshenev
'' Portions (c) 2008 Matthew E. Kimmel
''
'' Licensed under MIT license (details moved to EOF)
''
'' Modified by Jeff Ledger for IRkeyboard and 40x27 text
'' Requires BST (Eliminate unused Spin methods) to compile.

CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  IRpin = 5
  TVpin = 12
  SDpin = 0

DAT
'  storyFile byte "hhgtg.dat", 0
  storyFile byte "zork1.dat", 0

  saveFile byte "savegame.bin", 0

CON
  ' Dynamic memory size in bytes.
  ' This needs to be not less than game's dynamic memory block which varies.
  ' Zork I requires 11900 bytes
  MEM_SIZE=11900  
  ' Static virtual memory cached page size in bytes.
  ' Game files exceeding MEM_SIZE will use the cache. 
  ' Any number will work. Larger is faster.
  CACHE_SIZE=2000
  ' Routine and call stack size in words. This needs to be big enough for the game to run.
  ' Zork I works OK with 500 words. 
  STACK_SIZE=500
  ' Whether to display status bar
  SHOW_STATUS=true

 
OBJ


  text : "TV_Text_Half_Height"
  kb   : "webtv_keyboard"
  sd   : "fsrw.spin"


PUB start

  kb.start(IRpin) 'for WebTVIRKeyboard
  text.start(TVpin,0,1,40,27)

  text.out($0B)
  text.out(text#maxrows-1)

  sd.mount(SDpin)

  text.str(string("+++++++ ProZ-Machine +++++++", 13))
  loadStory
  parseHeader
  runEmulator

'---------- Instructions, spec section 4

CON
  ' opcode forms
  OPCODE_LONG = 0 
  OPCODE_SHORT = 1
  OPCODE_VARIABLE = 2
  OPCODE_EXTENDED = 3
        
  ' operand count
  OP_0 = 0 
  OP_1 = 1
  OP_2 = 2
  OP_VAR = 3

  ' operand types
  OPTYPE_LARGECONSTANT = 0 
  OPTYPE_SMALLCONSTANT = 1
  OPTYPE_VARIABLE = 2
  OPTYPE_OMITTED = 3


'---------- Opcode decoding, spec sections 4, 14

VAR
  long pc
  byte instruction
  byte opCode
  byte opForm
  byte opCount
  byte opType1
  byte opType2
  byte opType3 
  byte opType4 
  long operand1
  long operand2
  long operand3
  long operand4
  word resultAddr
  long branch
  ' branchReversed is technically a boolean, but we use a long so that
  ' comparison with longs works correctly. This can be improved...
  long branchReversed
  long randomNumber  
  
PRI runEmulator
  pc := initialPC
  stackFrame := 0
  stackPointer := 0
  newPage  
  mainLoop


PRI mainLoop | instructionPC, opTypeByte

  repeat
    instructionPC := pc
    instruction := getByte(pc++)

    ' Spec section 4.3
    if (versionNumber =>5 and opCode == 190)
      opForm := OPCODE_EXTENDED
      opCount := OP_VAR
      opCode := getByte(pc++)
      opTypeByte := getByte(pc++)
      opType1 := (opTypeByte >> 6) & 3
      opType2 := (opTypeByte >> 4) & 3
      opType3 := (opTypeByte >> 2) & 3
      opType4 := opTypeByte & 3
      error(string("TODO extended"))

    elseif (((instruction >> 6) & 3) == 3)
      opForm := OPCODE_VARIABLE

      if (((instruction >> 5) & 1) == 0)
        opCount := OP_2
      else 
        opCount := OP_VAR

      opCode := instruction & $1F
      opTypeByte := getByte(pc++)
      opType1 := (opTypeByte >> 6) & 3
      opType2 := (opTypeByte >> 4) & 3
      opType3 := (opTypeByte >> 2) & 3
      opType4 := opTypeByte & 3
      ' $c1 (dec 193) seems to be an exception from spec in 4.3.3
      ' and can have more operands
      if (opCount==OP_2 and instruction <> $c1)
        operand1 := getOperand(opType1)
        operand2 := getOperand(opType2)
        opType3 := OPTYPE_OMITTED
        opType4 := OPTYPE_OMITTED
      else 
        ' var operands
        operand1 := getOperand(opType1)
        operand2 := getOperand(opType2)
        operand3 := getOperand(opType3)
        operand4 := getOperand(opType4)
    elseif (((instruction >> 6) & 3) == 2)
      opForm := OPCODE_SHORT
      opType1 := (instruction >> 4) & 3
      if (opType1 == 3) 
        opCount := OP_0
      else 
        opCount := OP_1
      operand1 := getOperand(opType1)
      opCode := instruction & $0F

    else
      opForm := OPCODE_LONG
      opCount := OP_2
      if (((instruction >> 6) & 1) == 0)
        opType1 := OPTYPE_SMALLCONSTANT
      else 
        opType1 := OPTYPE_VARIABLE
      if (((instruction >> 5) & 1) == 0)
        opType2 := OPTYPE_SMALLCONSTANT
      else 
        opType2 := OPTYPE_VARIABLE
      operand1 := getOperand(opType1)
      operand2 := getOperand(opType2)
      opCode := instruction & $1F

    logDec(stepNumber++)
    logChar(" ")
    log(string("["))
    logDec(instructionPC)
    logChar("]")
    logChar("=")
    logDec(opCount)
    logChar(":")
    logDec(instruction)
    logChar(" ")
    logDec(opCode)
    logChar(" ")
    logDec(opForm)

    if (opCount == OP_1)
      logChar(" ")
      logDec(opType1)
      logChar("=")
      logDec(operand1)
    elseif ((opCount == OP_2 and instruction <> $c1))
      logChar(" ")
      logDec(opType1)
      logChar("=")
      logDec(operand1)
      logChar(" ")
      logDec(opType2)
      logChar("=")
      logDec(operand2)
    elseif (opCount ==OP_VAR or instruction == $c1)                
      logChar(" ")
      logDec(opType1)
      logChar("=")
      logDec(operand1)
      logChar(" ")
      logDec(opType2)
      logChar("=")
      logDec(operand2)
      logChar(" ")
      logDec(opType3)
      logChar("=")
      logDec(operand3)
      logChar(" ")
      logDec(opType4)
      logChar("=")
      logDec(operand4)


    if (opCount == OP_0)
        log(string(" OP_0 ")) 
        case(opCode) 
          $00:
            log(string("rtrue"))
            _rtrue        
          $01:
            log(string("rfalse"))
            _rfalse            
          $02:
            log(string("print"))
            _print            
          $03: 
            log(string("print_ret"))
           _print_ret            
          $04 : 
            log(string("nop"))
            ' do nothing            
          $05:
            log(string("save")) 
            getBranch
            _save            
          $06: 
            log(string("restore")) 
            getBranch
           _restore            
          $07: 
            log(string("restart")) 
             _restart            
          $08: 
            log(string("ret_popped"))
           _ret_popped                        
          $0a : 
            log(string("quit"))
           _quit                        
          $0b : 
            log(string("new_line"))
           _new_line                        
          other:
            log(string("Unknown opcode OP_0 "))
            logDec(opCode)
            log(string(" instruction "))
            logDec(instruction)
            error(string("Unknown opcode"))
                
    elseif (opCount == OP_1)
      log(string(" OP_1 "))  
      case(opCode) 
        $00: 
          log(string("jz")) 
          getBranch
          _jz                        
        $01: 
          log(string("get_sibling")) 
          getResult
          getBranch
          _get_sibling                        
        $02:
          log(string("get_child"))  
          getResult
          getBranch
          _get_child                        
        $03:
          log(string("get_parent"))   
          getResult
          _get_parent                        
        $04:
          log(string("get_prop_len"))   
          getResult
          _get_prop_len                        
        $05: 
          log(string("inc")) 
          _inc                        
        $06:
          log(string("dec")) 
          _dec                        
        $07:
          log(string("print_addr"))  
          _print_addr                        
        $0B: 
          log(string("ret")) 
          _ret                        
        $09:
          log(string("remove_obj")) 
          _remove_obj                        
        $0C:
          log(string("jump"))
          _jump                        
        $0A:
          log(string("print_obj")) 
          _print_obj                        
        $0D: 
          log(string("print_paddr")) 
          _print_paddr                        
        $0E: 
          log(string("load")) 
          getResult
          _load                        
        other:
          log(string("Unknown opcode OP_1 "))
          logDec(opCode)
          log(string(" instruction "))
          logDec(instruction)
          error(string("Unknown opcode"))
                
    elseif (opCount == OP_2)
      log(string(" OP_2 ")) 
      case(opCode) 
        $01:
          log(string("je"))
          getBranch
          _je                        
        $02: 
          log(string("jl"))
          getBranch
          _jl                        
        $03: 
          log(string("jg"))
          getBranch
          _jg                        
        $04: 
          log(string("dec_chk"))
          getBranch
          _dec_chk                        
        $05:
          log(string("inc_chk")) 
          getBranch
          _inc_chk                        
        $06: 
          log(string("jin")) 
          getBranch
          _jin                         
        $07: 
          log(string("test")) 
          getBranch
          _test                         
        $08: 
          log(string("or")) 
          getResult
          _or                         
        $09: 
          log(string("and")) 
          getResult
          _and                         
        $0A:
          log(string("test_attr"))  
          getBranch
          _test_attr                        
        $0B: 
          log(string("set_attr"))  
          _set_attr                        
        $0C: 
          log(string("clear_attr"))  
          _clear_attr                        
        $0D:
          log(string("store"))
          _store                        
        $0F:
          log(string("loadw")) 
          getResult
          _loadw                        
        $0E: 
          log(string("insert_obj")) 
          _insert_obj                        
        $10: 
          log(string("loadb"))
          getResult
          _loadb                        
        $11:
          log(string("get_prop"))   
          getResult
          _get_prop                        
        $12: 
          log(string("get_prop_addr"))   
          getResult
          _get_prop_addr                        
        $13: 
          log(string("get_next_prop"))   
          getResult
          _get_next_prop                        
        $14:
          log(string("add"))
          getResult
          _add                        
        $15: 
          log(string("sub"))
          getResult
          _sub                        
        $16: 
          log(string("mul"))
          getResult
          _mul                        
        $17: 
          log(string("div"))
          getResult
         _div                        
        $18: 
          log(string("mod"))
          getResult
          _mod                        
        other:
          log(string("Unknown opcode OP_2 "))
          logDec(opCode)
          log(string(" instruction "))
          logDec(instruction)
          error(string("Unknown opcode"))
                
    elseif (opCount == OP_VAR)
      log(string(" OP_VAR ")) 
      case(opCode) 
        $00:
          log(string("call"))
          _call                        
        $01: 
          log(string("storew"))
          _storew                        
        $02: 
          log(string("storeb"))
          _storeb                        
        $03: 
          log(string("put_prop"))
          _put_prop                        
        $04: 
          log(string("read"))
          if (versionNumber => 5)
            getResult
          _read                        
        $05:
          log(string("print_char"))
          _print_char
                         
        $06:
          log(string("print_num"))
          _print_num                        
        $07: 
          log(string("random"))
          getResult
          _random                        
        $08: 
          log(string("push"))
          _push                        
        $09: 
          log(string("pull"))
          _pull                         
        other:
          log(string("Unknown opcode OP_2 "))
          logDec(opCode)
          log(string(" instruction "))
          logDec(instruction)
          error(string("Unknown opcode"))

    logNL


PRI getOperand(operandType) 
  if (operandType == OPTYPE_OMITTED)
    return -1
  if (operandType ==  OPTYPE_VARIABLE)
    return getVariable(getByte(pc++))
  if (operandType ==  OPTYPE_SMALLCONSTANT)
    return  getByte(pc++)    
  if (operandType ==  OPTYPE_LARGECONSTANT)
    pc += 2
    return getWord(pc-2)    
  error(string("Invalid operandType"))

PRI getBranch | b1, b2, sval
  ' Get the first byte of the branch
  b1 := getByte(pc++)

  ' Check to see if logic is reversed
  if ((b1 & $80) == $80)
    branchReversed := false
    ' log(string(13, "Branch normal "))
  else
    branchReversed := true
    ' log(string(13, "Branch reversed "))

  ' If the branch is only one byte long, just set its
  ' value and return.
  if ((b1 & $40) == $40)
    branch := (b1 & $3f)
    log(string("short branch "))
    logDec(branch)
    logNL
    return

  ' Otherwise, construct a signed branch value.
  b2 := getByte(pc++)
  sval := (((((b1 & $3f) << 8) & $3f00) | b2) & $3fff)
  ' If the following makes no sense, see the Z-Machine spec
  ' on signed numbers.
  if ((sval & $2000) == $2000)
    branch := (sval - 16384)
  else 
    branch := sval

PRI doBranch
  if (branch == 0)
    log(string("rfalse to address="))
    logDec(stack[stackFrame])
    log(string(" stackFrame="))
    logDec(stackFrame)
    logNL
    _rfalse
  elseif (branch == 1)
    _rtrue
  else
     pc += (branch - 2)



PRI getResult
  resultAddr := getByte(pc++)
  log(string(13, "result address: "))
  logDec(resultAddr)
  return resultAddr


' ---------- Opcodes, spec section 15    

'' Signed 16-bit addition.
PRI _add | sop1, sop2, value
  sop1 := signedWord(operand1)
  sop2 := signedWord(operand2)
  value := unsignedWord(sop1 + sop2)
  setVariable(resultAddr, value)



PRI _and | value 
  value := operand1 & operand2
  setVariable(resultAddr, value)
    


PRI _call | i,oldStackFrame,numvars 

  ' Spec section 6.4.3
  if (operand1 == 0) 
    getResult
    setVariable(resultAddr,0)
    return

  oldStackFrame := stackFrame  
  stackFrame += STACK_FRAME_SIZE
  stack[stackFrame+STACK_FRAME_PREV_STACK_FRAME] := oldStackFrame
  ' store PC as two words: first low then high 
  stack[stackFrame+STACK_FRAME_PC] := pc & $ffff
  stack[stackFrame+STACK_FRAME_PC + 1] := pc >> 16
  stackPointer := stackFrame + STACK_FRAME_SIZE

  log(string("pc="))
  logdec(pc)
  log(string("stackFrame="))
  logdec(stackFrame)
  logNL

  if (versionNumber =<3)
    pc := operand1 * 2
  elseif (versionNumber =<5)
    pc := operand1 * 4
  else
    error(string("TODO _call: Unsupported version"))
  numvars := getByte(pc++)


  ' init local variables

  ' set all to 0 by default
  repeat i from 0 to 14  
    stack[stackFrame+STACK_FRAME_LOCAL_VARIABLES+i] := 0        

  ' before version 5 take default values from the code
  if (versionNumber < 5)
    log(string("numvars ")) 
    logDec(numvars) 
    log(string(" "))
    if numvars > 0 
      repeat i from 0 to numvars-1  
        stack[stackFrame+STACK_FRAME_LOCAL_VARIABLES+i] := getWord(pc + (i * 2))
            
    pc += numvars*2
        
 
  ' pass up to 3 arguments if provided
  if (opType2 <> OPTYPE_OMITTED)
    stack[stackFrame+STACK_FRAME_LOCAL_VARIABLES] := operand2
  if (opType3 <> OPTYPE_OMITTED)
    stack[stackFrame+STACK_FRAME_LOCAL_VARIABLES+1] := operand3
  if (opType4 <> OPTYPE_OMITTED)
    stack[stackFrame+STACK_FRAME_LOCAL_VARIABLES+2] := operand4



PRI _clear_attr
  clearAttribute(operand1,operand2)
    


PRI _dec | w 
  w := signedWord(getVariable(operand1))
  w := ((w - 1) // $10000)
  setVariable(operand1, w)
    


PRI _dec_chk 
  _dec
  operand1 := getVariable(operand1)
  _jl
    


PRI _div | sop1, sop2, value
  sop1 := signedWord(operand1)
  sop2 := signedWord(operand2)
  value := unsignedWord(sop1 / sop2)
  setVariable(resultAddr, value)



''  Get first object contained in given object, branching if this exists, i.e. is not nothing (i.e., is not 0).     
PRI _get_child | child
  child := getChild(operand1)
  setVariable(resultAddr,child)
  if ((child <> 0) <> branchReversed) 
    doBranch
        
    


''  Get parent object (note that this has no "branch if exists" clause).     
PRI _get_parent | parent
     
  parent := getParent(operand1)
  setVariable(resultAddr,parent)
    


''  Read property from object (resulting in the default value if it had no such declared property).
''  If the property has length 1, the value is only that byte. If it has length 2, the first two 
''  bytes of the property are taken as a word value. It is illegal for the opcode to be used 
''  if the property has length greater than 2, and the result is unspecified.     
PRI _get_prop | prop
  prop := getProperty(operand1,operand2)
  setVariable(resultAddr,prop)
    

PRI _get_prop_addr | addr
  addr := getPropertyAddress(operand1,operand2)
  setVariable(resultAddr,addr)
    


''  Get length of property data (in bytes) for the given object's property. 
''  It is illegal to try to find the property length of a property which does not 
''  exist for the given object, and an interpreter should halt with an error message 
''  (if it can efficiently check this condition).     
PRI _get_prop_len | len
  len := getPropertyLength(operand1)
  setVariable(resultAddr,len)
    


PRI _get_next_prop | num 
  num := getNextProperty(operand1,operand2)
  setVariable(resultAddr,num)
    


''  Get next object in tree, branching if this exists, i.e. is not 0.     
PRI _get_sibling | sib
  sib := getSibling(operand1)
  setVariable(resultAddr, sib)
  if ((sib <> 0) <> branchReversed) 
    doBranch
        

    

PRI _inc | w 
  w := signedWord(getVariable(operand1))
  w := ((w + 1) // $10000)
  setVariable(operand1, w)
    

PRI _inc_chk 
  _inc
  operand1 := getVariable(operand1)
  _jg
    

PRI _insert_obj
  insertObject(operand1,operand2)
    


PRI _je | sop1, sop2 
  ' je ($c1, dec 193 may take multiple operands)
  sop1 := signedWord(operand1)
  sop2 := signedWord(operand2)
  if (sop1==sop2)
    if (not branchReversed) 
      doBranch
      return
  elseif (instruction == $c1 and opType3 <> OPTYPE_OMITTED and sop1==signedWord(operand3))
    if (not branchReversed)
      doBranch
      return
  elseif (instruction == $c1 and opType4 <> OPTYPE_OMITTED and sop1==signedWord(operand4))
    if (not branchReversed)
      doBranch
      return
  else
    ' no matches
    if (branchReversed)
      doBranch
      return



PRI _jin | parent
  parent := getParent(operand1)
  if ((parent == operand2) <> branchReversed) 
    doBranch
        
    

PRI _jl | sop1, sop2 
  sop1 := signedWord(operand1)
  sop2 := signedWord(operand2)
  if ((sop1<sop2) <> branchReversed) 
    doBranch
        
    

PRI _jg | sop1, sop2  
  sop1 := signedWord(operand1)
  sop2 := signedWord(operand2)
  if ((sop1>sop2) <> branchReversed) 
    doBranch
        
    

PRI _jz 
  if ((operand1==0) <> branchReversed) 
    doBranch
        
    


PRI _jump 
  pc += (signedWord(operand1) - 2)
    


''  The value of the variable referred to by the operand is stored in the result.      
PRI _load 
  setVariable(resultAddr,getVariable(operand1))
    


''  Stores array->byte-index (i.e., the byte at address array+byte-index, 
''  which must lie in static or dynamic memory).
     
PRI _loadb 
  setVariable(resultAddr, getByte(operand1 + operand2))
    


PRI _loadw 
'  log("Address: " + (operand1 + 2 * operand2))
'  log("Result: " + getWord(operand1 + 2 * operand2))
  setVariable(resultAddr, getWord(operand1 + 2 * operand2))
    


PRI _mul | sop1, sop2, value   
  sop1 := signedWord(operand1)
  sop2 := signedWord(operand2)
  value := unsignedWord(sop1 * sop2)
  setVariable(resultAddr, value)
  log(string(13, "result: "))
  logDec(value)
    

PRI _mod | sop1, sop2, value 
  sop1 := signedWord(operand1)
  sop2 := signedWord(operand2)
  value := unsignedWord(sop1 // sop2)
  setVariable(resultAddr, value)
  log(string(13, "result: "))
  logDec(value)
    

PRI _new_line
  printChar(13)     


PRI _or | value 
  value := operand1 | operand2
  setVariable(resultAddr, value)
  log(string(13, "result: "))
  logDec(value)


''  Print the quoted (literal) Z-encoded string.     
PRI _print
  pc := printZString(pc) 


''  Print (Z-encoded) string at given byte address, in dynamic or static memory.    
PRI _print_addr 
  printZString(operand1)


''  Print ZSCII character. The operand must be a character code defined in ZSCII for output (see S 3). 
''  In particular, it must certainly not be negative or larger than 1023.    
PRI _print_char
  printChar(operand1) 


''  Print (signed) number in decimal.
PRI _print_num
  text.dec(signedWord(operand1))

    
PRI _print_obj
  printZString(getObjectName(operand1))
    

 
''  Print the quoted (literal) Z-encoded string, then pra new-line and then return true (i.e., 1).     
PRI _print_ret 
  _print
  _new_line
  _rtrue

''  Print the (Z-encoded) string at the given packed address in high memory.     
PRI _print_paddr | addr 
  addr := 0
  if (versionNumber =<3)
    addr := operand1 * 2
  elseif (versionNumber =<5)
    addr := operand1 * 4
  else
    error(string("TODO: Unsupported version"))

  printZString(addr)

'' Pulls value off a stack. (If the stack underflows, the interpreter should halt with
'' a suitable error message.)  In Version 6, the stack in question may be specified
'' as a user one: otherwise it is the game stack.    
PRI _pull 
  if ((versionNumber == 6) and (opType1<>OPTYPE_OMITTED)) 
    error(string("Unsupported version"))        
  setVariable(operand1,getVariable(0))
    


''  Pushes value onto the game stack.     
PRI _push 
  setVariable(0,operand1)
    


PRI _put_prop
  putProperty(operand1 ,operand2, operand3)


''  If range is positive, returns a uniformly random number between 1 and range. 
''  If range is negative, the random number generator is seeded to that value and the 
''  return value is 0. Most interpreters consider giving 0 as range illegal 
''  (because they attempt a division with remainder by the range), but correct 
''  behaviour is to reseed the generator in as random a way as the interpreter 
''  can (e.g. by using the time in milliseconds).   
PRI _random | rnd
  if (signedWord(operand1) < 0)
    randomNumber := operand1
  else 
    ' TODO this is not quite uniformly distributed
    rnd := unsignedWord(?randomNumber) // signedWord(operand1) 
    setVariable(resultAddr,rnd + 1)
    

PRI _read | input, curAddr, startToken, tokenLength, tokenEntryAddr, numTokens, stringLength 
  if (versionNumber < 4) 
    _show_status

  newPage

  curAddr := operand1 + 1

  ' first read the string into the buffer
  repeat
    input := kb.getKey
    ' check backspace
    if input == $C8 and curAddr > operand1 + 1 
      text.out($08)
      text.out(" ")
      text.out($08)
      curAddr--
      next
    
    ' ignore non-printables
    if input < 13 or input > 126    
      next
    text.out(input)    

    ' convert to lowercase
    if input => "A" and input =< "Z"
      input += constant("a" - "A") 

    ' store the character (except newline)
    if input <> 13       
      setByte(curAddr++, input)
  until input == 13

  ' terminator char
  setByte(curAddr, 0)

  ' length
  stringLength := curAddr - operand1 - 1     
  setByte(operand1, stringLength)

  ' if tokenize is not requested, we are done 
  if operand2 == 0
    return

  curAddr := operand1 + 1 
  startToken := operand1 + 1 
  tokenLength := 0
  tokenEntryAddr := operand2 + 2
  numTokens := 0

  ' now tokenize the string       
  repeat  
    input := getByte(curAddr)
    if input == " " or input == 0 
      if tokenLength <> 0 
        ' add last token
        addToken(startToken, tokenLength, operand1, tokenEntryAddr)
        tokenLength := 0
        numTokens++
        tokenEntryAddr += 4
    else
      if tokenLength==0 
        startToken := curAddr
      tokenLength++
    curAddr++
  until input == 0
 
  ' Finally, store the number of tokens tokenized in the parse buffer.
  setByte(operand2 + 1, numtokens)

''  Detach the object from its parent, so that it no longer has any parent.
''  (Its children remain in its possession.)     
PRI _remove_obj | parent
  parent := getParent(operand1)
  if (operand1 == 0)
    return 
  removeObject(parent,operand1)
    

PRI _restart 
  pc := initialPC
  if not loadStory
    error(string("reloading story"))
        
  stackPointer := 0
  stackFrame := 0
    

PRI _restore
  restoreState
  ' no need to branch. Either restore failed or we are now in a different game
    

PRI _ret
  ' pc is stored as two words
  pc := stack[stackFrame+STACK_FRAME_PC+1]
  pc := pc << 16 | stack[stackFrame+STACK_FRAME_PC]
  stackFrame := stack[stackFrame+STACK_FRAME_PREV_STACK_FRAME]
  stackPointer := stackFrame + STACK_FRAME_SIZE

  ' set result
  setVariable(getByte(pc++),operand1)
  


''  Pops top of stack and returns that.
'' (This is equivalent to ret sp, but is one byte cheaper.)     
PRI _ret_popped 
  operand1 := getVariable(0)
  _ret
    


''  Return false (i.e., 0) from the current routine.     
PRI _rfalse 
  operand1 := 0
  _ret


''  Return true (i.e., 1) from the current routine.     
PRI _rtrue 
  operand1 := 1
  _ret
    

PRI _save
  if (saveState<>branchReversed) 
    doBranch
        
    

PRI _set_attr
  setAttribute(operand1,operand2)
    

PRI _show_status | x, y
  if not SHOW_STATUS
    return

  ' Remember current position.
  ' Note: text driver does not report cursor location so we have
  ' to rely on our count
  x:=col
  y:=text#maxrows-1
  
  text.out($0C) ' set color
  text.out($04) ' color value
  text.out($01) ' home
  repeat text#maxcols
    text.out(" ")  
  text.out($01) ' home

  printStatus
  text.out($0A)  ' setX
  text.out(x)
  text.out($0B)  ' setY
  text.out(y) 
  text.out($0C) ' set color
  text.out($00)  ' default color  



PRI _store
  setVariable(operand1, operand2)


PRI _storew 
  setWord(operand1 + 2 * operand2, operand3)
    

PRI _storeb 
  setByte(operand1 + operand2, operand3 & $ff)
    

PRI _sub | sop1, sop2, value 
  sop1 := signedWord(operand1)
  sop2 := signedWord(operand2)
  value := unsignedWord(sop1 - sop2)
  setVariable(resultAddr, value)
  log(string(13, "result: "))
  logDec(value)
    
    

PRI _test 
  if ((operand1 & operand2) == operand2 <> branchReversed) 
    doBranch


PRI _test_attr
  if (hasAttribute(operand1,operand2) <> branchReversed) 
    doBranch


PRI _quit
  cogstop(cogid)


'---------- Stack and variables, spec section 6

CON
  STACK_FRAME_SIZE = 19                ' each stack frame is 19 words
  STACK_FRAME_PC = 0                   ' PC= two words
  STACK_FRAME_LOCAL_VARIABLES = 2      ' local vars = 14 words   
  STACK_FRAME_PREV_STACK_FRAME = 18    ' parent stack frame index: 1 word   

VAR
  word stackPointer, stackFrame
  ' combined call and routine stack for the game. 
  word stack[STACK_SIZE]  


PRI setVariable(variable, value) 
  if (variable==0)
    stack[stackPointer++] := value
  elseif (variable < $10) 
    ' local
    stack[stackFrame+constant(STACK_FRAME_LOCAL_VARIABLES-1)+variable] := value
  elseif (variable < $100)
    ' global
    setWord(globalVariableTable+(variable-$10)*2, value)
  else
    logDec(variable)
    logNL
    error(string("var address"))


PRI getVariable(variable):value  

  value := 0
  if (variable==0)
    value := stack[--stackPointer]
  elseif variable < $10
    ' local
    value := stack[stackFrame+constant(STACK_FRAME_LOCAL_VARIABLES-1)+variable]
  elseif variable < $100
    ' global
    value := getWord(globalVariableTable+(variable-$10)*2)
  else
    logDec(variable)
    logNL
    error(string("var address"))



PRI signedWord(w)
  if ((w & $8000) == $8000)
    return w - 65536
  else
    return w

PRI unsignedWord(w)
  w := w & $ffff
  if (w < 0)
    return 65536-(-w)
  else
    return w


'---------- Text, spec section 3

VAR
  byte currentAlphabet
  byte shift
  byte abbreviationType
  byte zsciiCode
  byte zsciiCharacter
  byte numberOfLines ' used for paging
  byte col ' used for paging

PRI newPage
  col:=0
  numberOfLines:=0   

PRI printChar(char)
  text.out(char)
  col++
  if col => text#maxcols or char==13 
    col:=0
    numberOfLines++
  if numberOfLines => text#maxrows - 1
    newPage
    text.str(string("--MORE--"))
    kb.getKey
    ' erase prompt
    repeat 8
      text.out(8)       
    repeat 8
      text.out(" ")       
    repeat 8
      text.out(8)       

PRI printZChar(code)| zString 
  if (zsciiCharacter>0)  ' ZSCII character part
    zsciiCode := (zsciiCode << 5) + code
    zsciiCharacter++
    if (zsciiCharacter == 3) 
      printChar(zsciiCode)
      zsciiCharacter := 0
      zsciiCode := 0
            
  elseif (abbreviationType > 0)  ' is this an abbreviation pointer?
    zString := 2*getWord(abbreviationTable + 64*(abbreviationType-1) + 2*code)
    printZString(zString)
    shift := 0
  elseif (code==0) 
    printChar(" ")
    shift := 0
  elseif (code==1) 
    if (versionNumber  == 1)
      printChar(13)
    else
      abbreviationType := 1
  elseif (code==2) 
    if (versionNumber < 3)
      shift := 1
    else
      abbreviationType:=2
  elseif (code==3) 
    if (versionNumber  < 3)
      shift := 2
    else
      abbreviationType:=3
  elseif (code==4) 
    if (versionNumber < 3)
      currentAlphabet := (currentAlphabet+1) // 3
    else
      shift:=1
  elseif (code==5) 
    if (versionNumber < 3)
      currentAlphabet := (currentAlphabet+2) // 3
    else
      shift:=2
  elseif (code=>6) 
    if ((currentAlphabet + shift) // 3 == 0)
      printChar(code+constant($61-6)) ' lowercase
    elseif ((currentAlphabet + shift) // 3 == 1) 
      printChar(code+constant($41-6)) ' uppercase
    elseif ((currentAlphabet + shift) // 3 == 2) 
      if (code == 6) 
        'ZSCII character follows
        zsciiCharacter := 1
        zsciiCode := 0
      else
        printChar(punctuationAlphabet[code-6])
                
    else 
      'should not happen
      error(string("Bad alphabet"))
            
    shift := 0
        
    

PRI printZString(pointer) | curWord, charCount

 ' reset shifts and shiftlocks (if any)
  shift := 0
  currentAlphabet := 0
  abbreviationType := 0
  zsciiCharacter := 0
  zsciiCode := 0

  repeat
    curWord := getWord(pointer)

    printZChar((curWord >> 10)&31)
    printZChar((curWord >> 5)&31)
    printZChar((curWord)&31)
    pointer+=2
  while ((curWord&$8000) == 0)   

  return pointer    

PRI printStatus | name, score, turns    
  name := getObjectName(getVariable(16))
  score := signedWord(getVariable(17))
  turns := signedWord(getVariable(18))
  printZString(name)
  text.str(string("  S:"))
  text.dec(score)
  text.str(string(" T:"))
  text.dec(turns)
  text.out(13)


' ---------- Dictionary, Spec section 13

PRI addToken(stringAddress, stringLength, inputBufferAddress, tokenAddress) | i, j, k, dictaddr, curaddr, dictentrysize, numdictentries, curzchar, numseparators, match, curindex, maxTokenLen, curword, n 
  ' log(string("addToken stringAddress="))
  ' logDec(stringAddress)
  ' log(string(" stringLength="))
  ' logDec(stringLength)
  ' log(string(" tokenAddress="))
  ' logDec(tokenAddress)
  ' log(string(" content=["))
  'repeat i from 0 to stringLength-1
  '    logChar(getByte(stringAddress+i))
  'log(string("]",13))
    
  dictaddr := dictionary
  numseparators := getByte(dictaddr)

  ' Set maximum token length, in words
  if (versionNumber < 4)
    maxtokenlen := 2
  else
    maxtokenlen := 3

  curaddr := dictaddr + 1 + numseparators
  dictentrysize := getByte(curaddr)
  curaddr++
  numdictentries := signedWord(getWord(curaddr))
  curaddr += 2
        if (numdictentries < 0)
            numdictentries := ||(numdictentries) ' We don't care whether the entries are sorted.

        ' Now, loop through the dictionary and check this token against each one
        ' in the dictionary.  Remember that curaddr points to the first dictionary
        ' entry at this point.
  repeat i from 0 to numdictentries-1
          '  Compare this dictionary entry against the encoded string.
    match := true
    curindex := 0
    repeat j from 0 to maxtokenlen-1
      curword := getWord(curaddr + (i * dictentrysize) + (j*2))
      repeat k from 2 to 0
        curzchar := ((curword >> (k*5)) & $1f)
          if (curindex == stringLength)
            if (curzchar <> 5)
              match := false
              quit
            else
              next
          else ' curindex valid
            ' TODO This is oversimplified and can only handle lowercase correctly 
             n := getByte(stringAddress+curIndex) - "a" + 6
             curindex++
             if (curzchar <> n) 
               match := false
               quit

        ' Break out of this token comparison if we know it's false.
        if ( not match)
          quit

    ' Break out of the dictionary walk if this token matches.
    if (match)
      quit

  ' This is a kludge.  It is possible for match to be true if
  ' the loop exits at the end of the dictionary.
  if (i => numdictentries)
    match := false

  ' log(string("Dic entry: "))          
  ' logDec(i)          
  ' log(string("/"))          
  ' logDec(numdictentries)          
  ' log(string("Match: "))          
  ' logDec(match)          
  ' logChar(13)          
  ' Store the token.  i still is the dictionary entry number of the matched entry.
  if (match) 
    setWord(tokenAddress,(curaddr + (i * dictentrysize))) ' Memory location of dictionary entry
    setByte(tokenAddress + 2,stringLength) ' Length of word
    setByte(tokenAddress + 3,stringAddress-inputBufferAddress) ' Position in input buffer 
  else 'if (bit == 0) ' If bit is set, leave the slot alone
    setWord(tokenAddress,0)
    setByte(tokenAddress + 2,stringLength) ' Length of word
    setByte(tokenAddress + 3,stringAddress-inputBufferAddress) ' Position in input buffer 


      
'---------- Header, spec section 11

VAR
  word initialPC
  word abbreviationTable
  word globalVariableTable
  word dictionary
  word objectTable
  word staticMemory
  byte versionNumber

PRI parseHeader 
  versionNumber := getByte($00)
  ' log(string("versionNumber "))
  ' logDec(versionNumber)
  ' logNL
  globalVariableTable :=getWord($0c)
  ' log(string("globalVariableTable "))
  ' logDec(globalVariableTable)
  ' logNL
  abbreviationTable := getWord($18)
  ' log(string("abbreviationTable "))
  ' logDec(abbreviationTable)
  ' logNL
  dictionary := getWord($08)
  ' log(string("dictionary "))
  ' logDec(dictionary)
  ' logNL
  objectTable := getWord($0a)
  ' log(string("objectTable "))
  ' logDec(objectTable)
  ' logNL
  staticMemory := getWord($0e)
  ' log(string("staticMemory "))
  ' logDec(staticMemory)
  ' logNL
  initialPC := getWord($06)
  ' log(string("initialPC "))
  ' logDec(initialPC)
  ' logNL




'---------- Object Table, spec section 12
' This section is adapted from Zax with minimal changes. 

' TODO should be version-dependent. These are defaults for V 3.
CON
  defaultsSize = 62
  objAttrSize = 4
  objHandleSize = 1
  objEntrySize = (objAttrSize + (3 * objHandleSize) + 2)


' Object manipulation routines

'' Return the sibling of an object
PRI getSibling(object) | sib  
  if (versionNumber =< 3)
    sib := getByte(objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + objHandleSize)
  else
    sib := getWord(objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + objHandleSize)

  return sib
    

'' Set the sibling of an object
PRI setSibling(object,sib)
    
  if (versionNumber =< 3)
    setByte((objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + objHandleSize),sib)
  else
    setWord((objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + objHandleSize),sib)
    

'' Return the first child of an object
PRI getChild(object) | child
    
  if (versionNumber =< 3)
    child := getByte(objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + (2 * objHandleSize))
  else
    child := getWord(objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + (2 * objHandleSize))

  return child
    

'' Set the child of an object
PRI setChild(object,child)
    
  if (versionNumber =< 3)
    setByte((objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + (2 * objHandleSize)),child)
  else
    setWord((objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + (2 * objHandleSize)),child)
    

'' Return an object's parent
PRI getParent(object) | parent

  if (versionNumber =< 3)
    parent := getByte(objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize)
  else
    parent := getWord(objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize)

  return parent
    

'' Set the parent of an object
PRI setParent(object,parent)
    
  if (versionNumber =< 3)
    setByte((objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize),parent)
  else
    setWord((objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize),parent)
    

'' Given its (non-zero) parent, remove an object from the
'' sibling chain.
PRI removeObject(parent,object) | curobject, prevobject

  ' It is legal for parent to be 0, in which case we just return.
  if (parent == 0)
    return

  curobject := getChild(parent)
  if (curobject == 0)
    error(string("object table"))
  if (curobject == object) 
    ' Remove the object
    setChild(parent,getSibling(object))
    setSibling(object,0)
    setParent(object,0)
    return
        

  ' Traverse the sibling chain until we find the object
  ' and its predecessor.
  prevobject := curobject
  curobject := getSibling(prevobject)
  repeat while ((curobject <> object) and (curobject <> 0)) 
    prevobject := curobject
    curobject := getSibling(prevobject)
        

  ' If we get here, curobject is either the object we're looking
  ' for or 0 (which is an error).
  if (curobject == 0)
    error(string("object table"))

  ' Remove the object from the chain, and set its sibling and parent to 0.
  setSibling(prevobject,getSibling(curobject))
  setSibling(object,0) ' Is this necessary?
  setParent(object,0)
    


'' Insert obj1 as obj2's first child.
PRI insertObject(obj1,obj2) | oldparent, oldfirst 

  ' First, remove the given object from its current
  ' position (if any).
  oldparent := getParent(obj1)
  if (oldparent > 0)
    removeObject(oldparent,obj1)

  ' Now insert it.
  oldfirst := getChild(obj2)
  setSibling(obj1,oldfirst)
  setChild(obj2,obj1)
  setParent(obj1,obj2)
    


'' Property manipulation routines

'' Return the length of the property starting at the given
'' byte address
PRI getPropertyLength(baddr) | b, length

  b := getByte(baddr-1)
  if (versionNumber < 4)
    length := (((b >> 5) & $07) + 1)
  else 
    if ((b & $80) == $80)
      length := (b & $3f)
    else
      length := (((b >> 6) & $01) + 1)
        
  return length
    

'' Get the address of the property list for the specified object.
PRI getPropertyList(object) | addr

  addr := getWord(objectTable + defaultsSize + ((object - 1) * objEntrySize) + objAttrSize + (3 * objHandleSize))
  return addr
    

'' Get the address of the specified property of the specified
'' object.  Return $0000 if there is no such property.
PRI getPropertyAddress(object,prop) | p, s, o, pnum, psize

  ' First, get the address of the property table for this
  ' object.
  p := getPropertyList(object)

  ' Now step through, looking for the specified property.
  ' Start by jumping over text header.
  o := getByte(p)
  p := p + (o * 2) + 1

  ' Now we're at the start of the property table.
  s := getByte(p)
  repeat while (s <> 0) 
            ' Get the property number and the size of this property.
    if (versionNumber < 4) 
      pnum := (s & $1f)
      psize := ((s >> 5) & $07) + 1
            
    else 
      pnum := (s & $3f)
      if ((s & $80) == $80) 
        p++
        psize := getByte(p)
        psize := psize & $3f
                
      else
        psize := ((s >> 6) & $03) + 1            

    ' Step over the size byte
    p++

    ' If this is the correct property, return its address
    ' otherwise, step over the property and loop.
    if (pnum == prop)
      return p
    else
      p := p + psize
    s := getByte(p)

  ' If we make it here, the property was not found.
  return 0
    

'' Get the first byte or word of a property--use the default
'' property if this one doesn't exist.
PRI getProperty(object,prop) | pdata

  ' Attempt to get the address of this property for this
  ' object.
  pdata := getPropertyAddress(object,prop)

  ' If the property exists, return it otherwise, return
  ' the appropriate value from the defaults table.
  if (pdata > 0) 
    if (getPropertyLength(pdata) == 1)
      return getByte(pdata)
    else
      return getWord(pdata)
        
  else
    return getWord(objectTable + ((prop - 1) * 2))
    

'' Return the property number of the property that follows
'' the specified property in the property list, or 0 if
'' the specified property doesn't exist.
PRI getNextProperty(object,prop) | propaddr, proplen, propnum

  ' First, if the property number is 0, just return the
  ' number of the first property of this object.
  if (prop == 0) 
    propaddr := getPropertyList(object)
    ' Skip over text header
    propaddr := propaddr + (getByte(propaddr) * 2)
    ' Return the number of the first property.
    ' This will work if the property number is 0, too.
    if (versionNumber =< 3)
      propnum := getByte(propaddr) & $1f
    else 
      propnum := getByte(propaddr) & $3f
    return propnum
        

  ' First, get the address of the specified property.
  ' If it doesn't exist, return 0.
  propaddr := getPropertyAddress(object,prop)
  if (propaddr == 0)
    return 0

  ' Now find out its length.
  proplen := getPropertyLength(propaddr)

  ' Skip over the property data
  propaddr += proplen

  ' Now return the number of the next property.  This will
  ' return 0 if the property is a 0 byte.
  ' Note: original Zax implementation has a bug here, handling verison 3 incorrectly
  if (versionNumber =< 3)     
    propnum := getByte(propaddr) & $1f
  else
    propnum := getByte(propaddr) & $3f
  return propnum
    

'' Return the address of the Z-String containing the specified
'' object's name.
PRI getObjectName(object) | addr

  addr := getPropertyList(object) + 1
  return addr
    

'' Put the specified value as the specified property of the
'' specified object.
PRI putProperty(object,prop,value) | propaddr, proplen

  ' First, get the address of this property.  Fail silently
  ' if the property does not exist.
  propaddr := getPropertyAddress(object,prop)
  if (propaddr == 0)
    return

  ' Now set the property, depending on its length.
  proplen := getPropertyLength(propaddr)
  if (proplen == 1)
    setByte(propaddr,(value & $ff))
  else
    setWord(propaddr,value)
    


' Attribute manipulation routines

'' Return true if the specified object has the specified
'' attribute otherwise return false.
PRI hasAttribute(object,attr) | whichbyte, whichbit, bitmask, attrbyte
  ' First, figure out which byte and bit we're looking at.
  whichbyte := attr / 8
  whichbit := attr // 8

  ' Flip the bit number around to something we can use in
  ' an AND.
  if whichbit <> 0
    bitmask := $80 >> whichbit
  else 
    bitmask := $80

  ' Now get the appropriate byte and test it.
  attrbyte := getByte(objectTable + defaultsSize + ((object - 1) * objEntrySize) + (whichbyte))

  if ((attrbyte & bitmask) == bitmask)
    return true
  else
    return false
    

'' Set an attribute on an object.
PRI setAttribute(object,attr) | whichbyte, whichbit, bitmask, attrbyte

  ' First, figure out which byte and bit we're looking at.
  whichbyte := attr / 8
  whichbit := attr // 8

  ' Flip the bit number around to something we can use in
  ' an OR.
  if whichbit <> 0
    bitmask := $80 >> whichbit
  else 
    bitmask := $80

  ' Now get the appropriate byte and set it.
  attrbyte := getByte(objectTable + defaultsSize + ((object - 1) * objEntrySize) + (whichbyte))
  attrbyte := attrbyte | bitmask
  setByte((objectTable + defaultsSize + ((object - 1) * objEntrySize) + (whichbyte)),attrbyte)
    

'' Clear an attribute on an object.
PRI clearAttribute(object,attr)| whichbyte, whichbit, bitmask, attrbyte

  ' First, figure out which byte and bit we're looking at.
  whichbyte := attr / 8
  whichbit := attr // 8

  ' Flip the bit number around to something we can use in
  ' a NOT.
  if whichbit <> 0
    bitmask := $80 >> whichbit
  else 
    bitmask := $80

  ' Now get the appropriate byte and clear it.
  attrbyte := getByte(objectTable + defaultsSize + ((object - 1) * objEntrySize) + (whichbyte))
  attrbyte := ((attrbyte & !bitmask) & $ff)
  setByte((objectTable + defaultsSize + ((object - 1) * objEntrySize) + (whichbyte)),attrbyte)


'---------- Save and restore

PRI saveState | i
  ' Only one file can be open at a time so we will close the story file 
  storyFileOpen := false 
  i := sd.popen(@saveFile, "w")
  if i<>0
    warning(string("open save file"))
    return false
  sd.pwrite(@pc, 4)
  sd.pwrite(@stackPointer, 2)
  sd.pwrite(@stackFrame, 2)
  sd.pwrite(@mem, staticMemory) 
  sd.pwrite(@stack, stackPointer*2) 
  sd.pclose
  return true  

PRI restoreState | i
  ' Only one file can be open at a time so we will close the story file 
  storyFileOpen := false 
  i := sd.popen(@saveFile, "r")
  if i<>0
    warning(string("open save file"))
    return false
  sd.pread(@pc, 4)
  sd.pread(@stackPointer, 2)
  sd.pread(@stackFrame, 2)
  sd.pread(@mem, staticMemory) 
  sd.pread(@stack, stackPointer*2) 
  sd.pclose
  return true  



'---------- Memory, spec section 1
VAR
   long cacheAddress 

PRI loadStory | i
  cacheAddress := $7fff_ffff  ' a number that is bigger than any file length  
  i := sd.popen(@storyFile, "r")
  if i<>0
    error(string("open story file"))
  
  i := sd.pread(@mem, MEM_SIZE)
  if i=<0
    error(string("read story file"))

  sd.pclose
  storyFileOpen := false  
  return true

'TODO range checking


' This is a simple implementation which works with story files in memory
'PRI getByte(address)
'  return mem[address] & $ff
'

var
  byte storyFileOpen

' This implementation uses virtual memory for static content  
' On a small device the full game file won't fit in memory so we
' have to load one page at a time.
PRI getByte(address) | i 
  if (address < MEM_SIZE)
    ' dynamic mem
    return mem[address] & $ff


  ' check cache first
  if (cacheAddress =< address and cacheAddress+CACHE_SIZE > address)
    return cache[address-cacheAddress] & $ff

  ' load new page into cache

  ' read file from start if we are going back
  if (address < cacheAddress or not storyFileOpen) 
    i := sd.popen(@storyFile, "r")
    storyFileOpen := true          
    if i<>0
      error(string("open story"))                
    cacheAddress := -CACHE_SIZE
    ' text.out("+")

  repeat
    i := sd.pread(@cache, CACHE_SIZE)
    cacheAddress += CACHE_SIZE
    ' text.out("*")
  while (not(cacheAddress =< address and cacheAddress+CACHE_SIZE > address))

  return cache[address-cacheAddress] & $ff


PRI setByte(address,value)
  mem[address] := value & $ff


PRI getWord(address)
  return (((getByte(address) << 8) | (getByte(address+1) & $ff)) & $ffff) 

PRI setWord(address, value)
  setByte(address, (value >> 8) & $ff)
  setByte(address+1, value & $ff)
    

' ---------- Logging
CON  
  debug = false
VAR  
  long stepNumber


PRI log(s)
'  if debug
'    term.str(s)
    

PRI logDec(number)
'  if debug
'    term.dec(number)  

PRI logChar(char)
'  if debug
'    term.char(char)

PRI logNL
  logChar(13)  

PRI warning(s) | errorStr
  errorStr := string("ERROR ") 
  logNL
  log(errorStr)
  log(s)
  logNL
  text.str(errorStr)
  text.str(s)
  text.out(13)

PRI error(s)
  warning(s)  
  ' This is a fatal error. Stop.
  cogstop(cogid)  

DAT

  punctuationAlphabet byte
    byte 0, 13, "0", "1", "2", "3", "4", "5", "6"    
    byte "7", "8", "9", ".", ",", "!", "?", "_", "#", "'", 34, "/", "\" 
    byte "-", ":", "(", ")"  

'  mem file "inform6-test.z5"
'  mem file "catseye.z3"
  mem byte 0 [MEM_SIZE]
  cache byte 0 [CACHE_SIZE]


''
'' Permission is hereby granted, free of charge, to any person obtaining
'' a copy of this software and associated documentation files
'' (the "Software"), to deal in the Software without restriction,
'' including without limitation the rights to use, copy, modify, merge,
'' publish, distribute, sublicense, and/or sell copies of the Software,
'' and to permit persons to whom the Software is furnished to do so,
'' subject to the following conditions:
''
'' The above copyright notice and this permission notice shall be included
'' in all copies or substantial portions of the Software.
''
'' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'' EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'' MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'' IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'' CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'' TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
